from datetime import datetime

import numpy as np
import reproject
from astropy.wcs import WCS
from ndcube import NDCube
from prefect import flow, task
from sunpy.coordinates import sun

from punchbowl.data.wcs import calculate_celestial_wcs_from_helio


@task
def reproject_array(input_array: np.ndarray, input_wcs: WCS, time: datetime,
                    output_wcs: WCS, output_shape: tuple) -> np.ndarray:
    """
    Core reprojection function.

    Core reprojection function of the PUNCH mosaic generation module.
        With an input data array and corresponding WCS object, the function
        performs a reprojection into the output WCS object system, along with
        a specified pixel size for the output array. This utilizes the adaptive
        reprojection routine implemented in the reprojection astropy package.

    Parameters
    ----------
    input_array
        input array to be reprojected
    input_wcs
        astropy WCS object describing the input array
    time
        time of the observation to reproject
    output_wcs
        astropy WCS object describing the coordinate system to transform to
    output_shape
        pixel shape of the reprojected output array


    Returns
    -------
    np.ndarray
        output array after reprojection of the input array


    Example Call
    ------------
    >>> reprojected_array = reproject_array(input_array, input_wcs, output_wcs, output_shape)

    """
    celestial_source = calculate_celestial_wcs_from_helio(input_wcs, time, output_shape)
    celestial_target = calculate_celestial_wcs_from_helio(output_wcs, time, output_shape)

    celestial_source.wcs.set_pv([(2, 1, (-sun.earth_distance(time) / sun.constants.radius).decompose().value)])

    return reproject.reproject_adaptive(
        (input_array, celestial_source), celestial_target, output_shape,
        roundtrip_coords=False, return_footprint=False,
    )


@flow(validate_parameters=False)
def reproject_many_flow(data: list[NDCube], trefoil_wcs: WCS, trefoil_shape: np.ndarray) -> list[NDCube]:
    """Reproject many flow."""
    data_result = [reproject_array.submit(d.data, d.wcs, d.meta.astropy_time, trefoil_wcs, trefoil_shape) for d in data]
    uncertainty_result = [reproject_array.submit(d.uncertainty.array, d.wcs,
                                                 d.meta.astropy_time, trefoil_wcs, trefoil_shape) for d in data]

    return [NDCube(data=data_result[i].result(),
                   uncertainty=uncertainty_result[i].result(),
                   wcs=trefoil_wcs,
                   meta=d.meta) for i, d in enumerate(data)]



import pathlib

from ndcube import NDCube
from prefect.testing.utilities import prefect_test_harness

from punchbowl.data.tests.test_io import sample_ndcube
from punchbowl.level1.flow import level1_core_flow

# def test_core_flow_runs_with_filenames(sample_punchdata):
#     with tempfile.TemporaryDirectory() as tmpdirname:
#         input_name = tmpdirname + "/test_input.fits"
#         output_name = tmpdirname + "/test_output.fits"
#         sample_punchdata(shape=(2048, 2048)).write(input_name)
#         with prefect_test_harness():
#             level1_core_flow(input_name, output_filename=output_name)
#         output = PUNCHData.from_fits(output_name)
#         assert isinstance(output, PUNCHData)

THIS_DIRECTORY = pathlib.Path(__file__).parent.resolve()


def test_core_flow_runs_with_objects_and_calibration_files(sample_ndcube):
    cube = sample_ndcube(shape=(10, 10), code="CR1", level="0")
    quartic_coefficient_path = THIS_DIRECTORY / "data" / "test_quartic_coeffs.fits"
    vignetting_path = THIS_DIRECTORY / "data" / "PUNCH_L1_GR1_20240222163425_v1.fits"

    with prefect_test_harness():
        output = level1_core_flow([cube],
                                  quartic_coefficient_path=quartic_coefficient_path,
                                  vignetting_function_path=vignetting_path)
    assert isinstance(output[0], NDCube)



import pytest
from astropy.wcs import WCS
from ndcube import NDCube

from punchbowl.data.tests.test_io import sample_ndcube
from punchbowl.level2.merge import merge_many_polarized_task


@pytest.fixture
def sample_data_list(sample_ndcube):
    """
    Generate a list of sample PUNCHData objects for testing
    """
    sample_pd1 = sample_ndcube((4096, 4096))
    sample_pd2 = sample_ndcube((4096, 4096))
    return [sample_pd1, sample_pd2]


@pytest.mark.prefect_test
def test_merge_many_task(sample_data_list):
    """
    Test the image_merge prefect flow using a test harness
    """
    trefoil_wcs = WCS("level2/data/trefoil_hdr.fits")
    trefoil_wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"  # TODO: figure out why this is necessary, seems like a bug

    pd_list = sample_data_list
    output_punchdata = merge_many_polarized_task.fn(pd_list, trefoil_wcs)
    assert isinstance(output_punchdata, NDCube)
    assert output_punchdata.data.shape == (3, 4096, 4096)




import numpy as np
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from ndcube import NDCube
from prefect import task

from punchbowl.data import NormalizedMetadata


@task
def merge_many_polarized_task(data: list[NDCube], trefoil_wcs: WCS) -> NDCube:
    """Merge many task and carefully combine uncertainties."""
    trefoil_data_layers, trefoil_uncertainty_layers = [], []
    for polarization in [-60, 0, 60]:
        selected_images = [d.data for d in data if d.meta["POLAR"].value == polarization]
        if len(selected_images) > 0:
            reprojected_data = np.stack(selected_images, axis=-1)
            reprojected_weights = np.stack([1/np.square(d.uncertainty.array)
                                            for d in data if d.meta["POLAR"].value == polarization],
                                               axis=-1)
            reprojected_weights[reprojected_weights <= 0] = 1E-8

            trefoil_data_layers.append(np.nansum(reprojected_data * reprojected_weights, axis=2) /
                                       np.nansum(reprojected_weights, axis=2))
            trefoil_uncertainty_layers.append(1/np.nansum(np.sqrt(reprojected_weights), axis=2))
        else:
            trefoil_data_layers.append(np.zeros((4096, 4096)))
            trefoil_uncertainty_layers.append(np.zeros((4096, 4096))-999)

    output_meta = NormalizedMetadata.load_template("PTM", "2")
    output_meta["DATE-OBS"] = data[0].meta["DATE-OBS"].value  # TODO: do this better and fill rest of meta

    return NDCube(
        data=np.stack(trefoil_data_layers, axis=0),
        uncertainty=StdDevUncertainty(np.stack(trefoil_uncertainty_layers, axis=0)),
        wcs=trefoil_wcs,
        meta=output_meta,
    )

@task
def merge_many_clear_task(data: list[NDCube], trefoil_wcs: WCS) -> NDCube:
    """Merge many task and carefully combine uncertainties."""
    trefoil_data_layers, trefoil_uncertainty_layers = [], []
    selected_images = [d.data for d in data]

    if len(selected_images) > 0:
        reprojected_data = np.stack(selected_images, axis=-1)
        reprojected_weights = np.stack([1/np.square(d.uncertainty.array) for d in data], axis=-1)
        reprojected_weights[reprojected_weights <= 0] = 1E-8

        trefoil_data_layers.append(np.nansum(reprojected_data * reprojected_weights, axis=2) /
                                   np.nansum(reprojected_weights, axis=2))
        trefoil_uncertainty_layers.append(1/np.nansum(np.sqrt(reprojected_weights), axis=2))
    else:
        trefoil_data_layers.append(np.zeros((4096, 4096)))
        trefoil_uncertainty_layers.append(np.zeros((4096, 4096))-999)

    output_meta = NormalizedMetadata.load_template("CTM", "Q")
    output_meta["DATE-OBS"] = data[0].meta["DATE-OBS"].value  # TODO: do this better and fill rest of meta

    return NDCube(
        data=np.stack(trefoil_data_layers, axis=0),
        uncertainty=StdDevUncertainty(np.stack(trefoil_uncertainty_layers, axis=0)),
        wcs=trefoil_wcs,
        meta=output_meta,
    )




import astropy.units as u
import solpolpy
from ndcube import NDCollection, NDCube
from prefect import get_run_logger, task


def resolve_polarization(data_list: list[NDCube]) -> list[NDCube]:
    """
    Take a set of input data in the camera MZP frame and convert to the solar MZP frame.

    Parameters
    ----------
    data_list : List[PUNCHData]
        List of PUNCHData objects on which to resolve polarization

    Returns
    -------
    List[PUNCHData]
        modified version of the input with polarization resolved

    """
    # Unpack data into a NDCollection object
    # TODO: don't assume the order and that there are only 3... that is not the case
    data_dictionary = dict(zip(["M", "Z", "P"], data_list, strict=False))
    input_collection = NDCollection(data_dictionary)
    data_collection = NDCollection({k: NDCube(data=input_collection[k].data,
                                 wcs=input_collection[k].wcs,
                                 meta={"POLAR": input_collection[k].meta["POLAR"].value * u.degree})
                       for k in ["M", "Z", "P"]})

    out = []
    resolved_data_collection = solpolpy.resolve(data_collection, "MZP", imax_effect=True)
    for key in resolved_data_collection:
        resolved_data_collection[key].meta = input_collection[key].meta
        resolved_data_collection[key].uncertainty = input_collection[key].uncertainty
        out.append(resolved_data_collection[key])

    return out


@task
def resolve_polarization_task(data_list: list[NDCube]) -> list[NDCube]:
    """
    Prefect task for polarization resolving.

    Parameters
    ----------
    data_list : List[PUNCHData]
        List of PUNCHData objects on which to resolve polarization

    Returns
    -------
    List[PUNCHData]
        modified version of the input with polarization resolved

    """
    logger = get_run_logger()
    logger.info("resolve_polarization started")

    # Resolve polarization
    data_list = resolve_polarization(data_list)

    logger.info("resolve_polarization ended")

    for data_object in data_list:
        data_object.meta.history.add_now("LEVEL2-resolve_polarization", "polarization resolved")
    return data_list



import numpy as np
import pytest
from astropy.time import Time
from astropy.wcs import WCS
from prefect.testing.utilities import prefect_test_harness

from punchbowl.data.tests.test_io import sample_ndcube
from punchbowl.level2.resample import reproject_array, reproject_many_flow


@pytest.fixture
def sample_punchdata_list(sample_ndcube):
    """
    Generate a list of sample PUNCHData objects for testing
    """
    sample_pd1 = sample_ndcube((50, 50))
    sample_pd2 = sample_ndcube((50, 50))
    return [sample_pd1, sample_pd2]


@pytest.fixture
def sample_wcs() -> WCS:
    """Generate a sample WCS for testing."""
    def _sample_wcs(naxis=2, crpix=(0, 0), crval=(0, 0), cdelt=(1, 1),
                    ctype=("HPLN-ARC", "HPLT-ARC")):
        generated_wcs = WCS(naxis=naxis)

        generated_wcs.wcs.crpix = crpix
        generated_wcs.wcs.crval = crval
        generated_wcs.wcs.cdelt = cdelt
        generated_wcs.wcs.ctype = ctype

        return generated_wcs
    return _sample_wcs


@pytest.mark.parametrize("crpix, crval, cdelt",
                         [((0, 0), (0, 0), (1, 1)),
                          ((0, 0), (1, 1), (1, 1)),
                          ((1, 1), (2, 2), (1, 1))])
def test_reproject_array(sample_wcs, crpix, crval, cdelt, output_shape=(50, 50)):
    """
    Test reproject_array usage
    """
    shape = (50, 50)
    test_wcs = sample_wcs(crpix=crpix, crval=crval, cdelt=cdelt)
    expected = np.random.random(shape)
    now = Time.now()
    actual = reproject_array.fn(expected, test_wcs, now, test_wcs, output_shape)

    assert actual.shape == expected.shape


def test_reproject_many_flow(sample_punchdata_list):
    trefoil_wcs = WCS("level2/data/trefoil_hdr.fits")
    trefoil_wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"  # TODO: figure out why this is necessary, seems like a bug

    trefoil_shape = (128, 128)
    with prefect_test_harness():
        output = reproject_many_flow(sample_punchdata_list, trefoil_wcs, trefoil_shape)
    for result in output:
        assert result.data.shape == trefoil_shape
        assert result.uncertainty.array.shape == trefoil_shape



from punchbowl.data.history import History
from punchbowl.data.io import get_base_file_name, load_ndcube_from_fits, write_ndcube_to_fits
from punchbowl.data.meta import NormalizedMetadata
from punchbowl.data.wcs import load_trefoil_wcs

__all__ = ["History",
           "NormalizedMetadata",
           "load_ndcube_from_fits",
           "write_ndcube_to_fits",
           "get_base_file_name",
           "load_trefoil_wcs"]



# import tempfile
#
# from prefect.testing.utilities import prefect_test_harness
#
# from punchbowl.data import PUNCHData
# from punchbowl.level2.flow import level2_core_flow
# from punchbowl.tests.test_data import sample_punchdata, sample_data_random
#

# def test_core_flow_runs_with_filenames(sample_punchdata):
#     with tempfile.TemporaryDirectory() as tmpdirname:
#         input_name = tmpdirname + "/test_input.fits"
#         output_name = tmpdirname + "/test_output.fits"
#         sample_punchdata(shape=(2048, 2048)).write(input_name)
#         with prefect_test_harness():
#             level1_core_flow(input_name, output_filename=output_name)
#         output = PUNCHData.from_fits(output_name)
#         assert isinstance(output, PUNCHData)
#         # todo: test more things


# def test_core_flow_runs_with_objects(sample_punchdata):
#     with prefect_test_harness():
#         sample_list = [sample_punchdata(shape=(2048, 2048)),
#                        sample_punchdata(shape=(2048, 2048)),
#                        sample_punchdata(shape=(2048, 2048)),
#                        sample_punchdata(shape=(2048, 2048)),
#                        sample_punchdata(shape=(2048, 2048))]
#         output = level2_core_flow(sample_list)
#     assert isinstance(output[0], PUNCHData)
#     # todo: test more things


# TODO: reactivate



from astropy.nddata import StdDevUncertainty
from ndcube import NDCube
from prefect import flow, get_run_logger

from punchbowl.data import get_base_file_name, load_trefoil_wcs
from punchbowl.data.meta import NormalizedMetadata
from punchbowl.data.wcs import load_quickpunch_mosaic_wcs, load_quickpunch_nfi_wcs
from punchbowl.exceptions import IncorrectFileCountError
from punchbowl.level2.bright_structure import identify_bright_structures_task
from punchbowl.level2.merge import merge_many_clear_task, merge_many_polarized_task
from punchbowl.level2.polarization import resolve_polarization_task
from punchbowl.level2.resample import reproject_many_flow
from punchbowl.util import load_image_task, output_image_task

ORDER = ["PM1", "PZ1", "PP1",
         "PM2", "PZ2", "PP2",
         "PM3", "PZ3", "PP3",
         "PM4", "PZ4", "PP4"]

ORDER_QP = ["CR1", "CR2", "CR3", "CR4"]


@flow(validate_parameters=False)
def level2_core_flow(data_list: list[str] | list[NDCube],
                     voter_filenames: list[list[str]],
                     output_filename: str | None = None) -> list[NDCube]:
    """Level 2 core flow."""
    logger = get_run_logger()
    logger.info("beginning level 2 core flow")
    if len(data_list) != 12:
        msg = f"Received {len(data_list)} files when need 12."
        raise IncorrectFileCountError(msg)

    data_list = [load_image_task(d) if isinstance(d, str) else d for d in data_list]

    # order the data list so it can be processed properly
    ordered_data_list = []
    for order_element in ORDER:
        found = False
        for data_element in data_list:
            typecode = data_element.meta["TYPECODE"].value
            obscode = data_element.meta["OBSCODE"].value
            if typecode == order_element[:2] and obscode == order_element[2]:
                ordered_data_list.append(data_element)
                found = True
        if not found:
            msg = f"Did not receive {order_element} file."
            raise IncorrectFileCountError(msg)
    logger.info(f"Ordered files are {[get_base_file_name(cube) for cube in ordered_data_list]}")

    trefoil_wcs, trefoil_shape = load_trefoil_wcs()

    # TODO make more robust... fails if a file is missing or out of order right now
    data_list = (resolve_polarization_task(ordered_data_list[:3]) + resolve_polarization_task(ordered_data_list[3:6])
                 + resolve_polarization_task(ordered_data_list[6:9]) + resolve_polarization_task(ordered_data_list[9:]))
    data_list = reproject_many_flow(data_list, trefoil_wcs, trefoil_shape)
    data_list = [identify_bright_structures_task(cube, voter_filenames)
                 for cube, voter_filenames in zip(data_list, voter_filenames, strict=True)]
    output_data = merge_many_polarized_task(data_list, trefoil_wcs)

    if output_filename is not None:
        output_image_task(output_data, output_filename)

    logger.info("ending level 2 core flow")
    return [output_data]


@flow(validate_parameters=False)
def levelq_core_flow(data_list: list[str] | list[NDCube],
                     output_filename: list[str] | None = None) -> list[NDCube]:
    """Level quickPUNCH core flow."""
    logger = get_run_logger()
    logger.info("beginning level quickPUNCH core flow")
    if len(data_list) != 4:
        msg = f"Received {len(data_list)} files when need 4."
        raise IncorrectFileCountError(msg)

    data_list = [load_image_task(d) if isinstance(d, str) else d for d in data_list]

    # order the data list so it can be processed properly
    ordered_data_list = []
    for order_element in ORDER_QP:
        found = False
        for data_element in data_list:
            typecode = data_element.meta["TYPECODE"].value
            obscode = data_element.meta["OBSCODE"].value
            if typecode == order_element[:2] and obscode == order_element[2]:
                ordered_data_list.append(data_element)
                found = True
        if not found:
            msg = f"Did not receive {order_element} file."
            raise IncorrectFileCountError(msg)
    logger.info(f"Ordered files are {[get_base_file_name(cube) for cube in ordered_data_list]}")

    quickpunch_mosaic_wcs, quickpunch_mosaic_shape = load_quickpunch_mosaic_wcs()
    quickpunch_nfi_wcs, quickpunch_nfi_shape = load_quickpunch_nfi_wcs()

    data_list_mosaic = reproject_many_flow(ordered_data_list, quickpunch_mosaic_wcs, quickpunch_mosaic_shape)
    data_list_nfi = reproject_many_flow(ordered_data_list[-1:], quickpunch_nfi_wcs, quickpunch_nfi_shape)

    output_data_mosaic = merge_many_clear_task(data_list_mosaic, quickpunch_mosaic_wcs)

    output_meta_nfi = NormalizedMetadata.load_template("CNN", "Q")
    output_meta_nfi["DATE-OBS"] = data_list_nfi[0].meta["DATE-OBS"].value
    output_data_nfi =  NDCube(
        data=data_list_nfi[0].data,
        uncertainty=StdDevUncertainty(data_list_nfi[0].uncertainty.array),
        wcs=quickpunch_nfi_wcs,
        meta=output_meta_nfi,
        )

    if output_filename is not None:
        output_image_task(output_data_mosaic, output_filename[0])
        output_image_task(output_data_nfi, output_filename[1])

    logger.info("ending level quickPUNCH core flow")
    return [output_data_mosaic, output_data_nfi]



from pathlib import Path
from collections.abc import Callable

import astropy.units as u
import numpy as np
from ndcube import NDCube
from prefect import flow, get_run_logger
from regularizepsf import ArrayCorrector, CoordinatePatchCollection, simple_psf

from punchbowl.data import NormalizedMetadata
from punchbowl.data.units import dn_to_msb
from punchbowl.level1.alignment import align_task
from punchbowl.level1.deficient_pixel import remove_deficient_pixels_task
from punchbowl.level1.despike import despike_task
from punchbowl.level1.destreak import destreak_task
from punchbowl.level1.initial_uncertainty import update_initial_uncertainty_task
from punchbowl.level1.psf import correct_psf_task
from punchbowl.level1.quartic_fit import perform_quartic_fit_task
from punchbowl.level1.sqrt import decode_sqrt_data
from punchbowl.level1.stray_light import remove_stray_light_task
from punchbowl.level1.vignette import correct_vignetting_task
from punchbowl.util import load_image_task, output_image_task


@flow(validate_parameters=False)
def generate_psf_model_core_flow(input_filepaths: [str],
                                 psf_size: int = 32,
                                 patch_size: int = 256,
                                 target_fwhm: float = 3.25) -> ArrayCorrector:
    """Generate PSF model."""
    # Define the target PSF as a symmetric Gaussian
    center = patch_size / 2
    sigma = target_fwhm / 2.355

    @simple_psf
    def target(x, y, x0=center, y0=center, sigma_x=sigma, sigma_y=sigma) -> np.ndarray:  # noqa: ANN001
        return np.exp(-(np.square(x - x0) / (2 * np.square(sigma_x)) + np.square(y - y0) / (2 * np.square(sigma_y))))

    target_evaluation = target(*np.meshgrid(np.arange(patch_size), np.arange(patch_size)))

    # Build the PSF model using the target PSF
    coordinate_patch_collection = CoordinatePatchCollection.find_stars_and_average(input_filepaths,
                                                                                   psf_size,
                                                                                   patch_size)
    return coordinate_patch_collection.to_array_corrector(target_evaluation)


@flow(validate_parameters=False)
def level1_core_flow(
    input_data: list[str] | list[NDCube],
    gain: float = 4.9,
    bias_level: float = 100,
    dark_level: float = 55.81,
    read_noise_level: float = 17,
    bitrate_signal: int = 16,
    quartic_coefficient_path: str | None = None,
    despike_unsharp_size: int = 1,
    despike_method: str = "median",
    despike_alpha: float = 3.0,
    despike_dilation: int = 0,
    exposure_time: float = 49 * 1000,
    readout_line_time: float = 163/2148,
    reset_line_time: float = 163/2148,
    vignetting_function_path: str | None = None,
    stray_light_path: str | None = None,
    deficient_pixel_map_path: str | None = None,
    deficient_pixel_method: str = "median",
    deficient_pixel_required_good_count: int = 3,
    deficient_pixel_max_window_size: int = 10,
    psf_model_path: str | None = None,
    alignment_mask: Callable | None = None,
    output_filename: str | None = None,
) -> list[NDCube]:
    """Core flow for level 1."""
    logger = get_run_logger()

    logger.info("beginning level 1 core flow")

    output_data = []
    for i, this_data in enumerate(input_data):
        data = load_image_task(this_data) if isinstance(this_data, str) else this_data

        data = decode_sqrt_data(data)
        data = update_initial_uncertainty_task(data,
                                               bias_level=bias_level,
                                               dark_level=dark_level,
                                               gain=gain,
                                               read_noise_level=read_noise_level,
                                               bitrate_signal=bitrate_signal,
                                               )
        data = perform_quartic_fit_task(data, quartic_coefficient_path)

        if data.meta["OBSCODE"].value == "4":
            scaling = {"gain": 4.9 * u.photon / u.DN,
                       "wavelength": 530. * u.nm,
                       "exposure": 49 * u.s,
                       "aperture": 49.57 * u.mm ** 2}
        else:
            scaling = {"gain": 4.9 * u.photon / u.DN,
                       "wavelength": 530. * u.nm,
                       "exposure": 49 * u.s,
                       "aperture": 34 * u.mm ** 2}
        data.data[:, :] = np.clip(dn_to_msb(data.data[:, :], data.wcs, **scaling), a_min=0, a_max=None)
        data.uncertainty.array[:, :] = dn_to_msb(data.uncertainty.array[:, :], data.wcs, **scaling)

        data = despike_task(data,
                            unsharp_size=despike_unsharp_size,
                            method=despike_method,
                            alpha=despike_alpha,
                            dilation=despike_dilation)
        data = destreak_task(data,
                             exposure_time=exposure_time,
                             reset_line_time=reset_line_time,
                             readout_line_time=readout_line_time)
        data = correct_vignetting_task(data, Path(vignetting_function_path))
        data = remove_deficient_pixels_task(data,
                                            deficient_pixel_map_path,
                                            required_good_count=deficient_pixel_required_good_count,
                                            max_window_size=deficient_pixel_max_window_size,
                                            method=deficient_pixel_method)
        data = remove_stray_light_task(data, stray_light_path)
        data = correct_psf_task(data, psf_model_path)

        # set up alignment mask
        observatory = int(data.meta["OBSCODE"].value)
        if observatory < 4:
            alignment_mask = lambda x, y: (x > 100) * (x < 1900) * (y > 250) * (y < 1900)  # noqa: E731
        else:
            alignment_mask = lambda x, y: (((x < 824) + (x > 1224)) * ((y < 824) + (y > 1224))  # noqa: E731
                                           * (x > 100) * (x < 1900) * (y > 100) * (y < 1900))
        data = align_task(data, mask=alignment_mask)

        # Repackage data with proper metadata
        product_code = data.meta["TYPECODE"].value + data.meta["OBSCODE"].value
        new_meta = NormalizedMetadata.load_template(product_code, "1")
        new_meta["DATE-OBS"] = data.meta["DATE-OBS"].value  # TODO: do this better and fill rest of meta

        output_header = new_meta.to_fits_header(data.wcs)
        for key in output_header:
            if (key in data.meta) and output_header[key] == "" and (key != "COMMENT") and (key != "HISTORY"):
                new_meta[key].value = data.meta[key].value

        data = NDCube(data=data.data, meta=new_meta, wcs=data.wcs, unit=data.unit, uncertainty=data.uncertainty)

        if output_filename is not None and i < len(output_filename) and output_filename[i] is not None:
            output_image_task(data, output_filename[i])
        output_data.append(data)
        logger.info("ending level 1 core flow")
    return output_data


@flow(validate_parameters=False)
def level05_core_flow(
    input_data: list[str] | list[NDCube],
    gain: float = 4.9,
    bias_level: float = 100,
    dark_level: float = 55.81,
    read_noise_level: float = 17,
    bitrate_signal: int = 16,
    quartic_coefficient_path: str | None = None,
    exposure_time: float = 49 * 1000,
    readout_line_time: float = 163/2148,
    reset_line_time: float = 163/2148,
    output_filename: str | None = None,
) -> list[NDCube]:
    """Core flow for level 0.5."""
    logger = get_run_logger()

    logger.info("beginning level 0.5 core flow")

    output_data = []
    for i, this_data in enumerate(input_data):
        data = load_image_task(this_data) if isinstance(this_data, str) else this_data
        data = decode_sqrt_data(data)
        data = update_initial_uncertainty_task(data,
                                               bias_level=bias_level,
                                               dark_level=dark_level,
                                               gain=gain,
                                               read_noise_level=read_noise_level,
                                               bitrate_signal=bitrate_signal,
                                               )
        data = perform_quartic_fit_task(data, quartic_coefficient_path)

        if data.meta["OBSCODE"].value == "4":
            scaling = {"gain": 4.9 * u.photon / u.DN,
                       "wavelength": 530. * u.nm,
                       "exposure": 49 * u.s,
                       "aperture": 49.57 * u.mm ** 2}
        else:
            scaling = {"gain": 4.9 * u.photon / u.DN,
                       "wavelength": 530. * u.nm,
                       "exposure": 49 * u.s,
                       "aperture": 34 * u.mm ** 2}
        data.data[:, :] = np.clip(dn_to_msb(data.data[:, :], data.wcs, **scaling), a_min=0, a_max=None)
        data.uncertainty.array[:, :] = dn_to_msb(data.uncertainty.array[:, :], data.wcs, **scaling)

        data = destreak_task(data,
                             exposure_time=exposure_time,
                             reset_line_time=reset_line_time,
                             readout_line_time=readout_line_time)

        # set up alignment mask
        observatory = int(data.meta["OBSCODE"].value)
        if observatory < 4:
            alignment_mask = lambda x, y: (x > 100) * (x < 1900) * (y > 250) * (y < 1900)  # noqa: E731
        else:
            alignment_mask = lambda x, y: (((x < 824) + (x > 1224)) * ((y < 824) + (y > 1224))  # noqa: E731
                                           * (x > 100) * (x < 1900) * (y > 100) * (y < 1900))

        data = align_task(data, mask=alignment_mask)

        # Repackage data with proper metadata
        product_code = data.meta["TYPECODE"].value + data.meta["OBSCODE"].value
        new_meta = NormalizedMetadata.load_template(product_code, "1")
        new_meta["DATE-OBS"] = data.meta["DATE-OBS"].value  # TODO: do this better and fill rest of meta

        output_header = new_meta.to_fits_header(data.wcs)
        for key in output_header:
            if (key in data.meta) and output_header[key] == "" and (key != "COMMENT") and (key != "HISTORY"):
                new_meta[key].value = data.meta[key].value

        data = NDCube(data=data.data, meta=new_meta, wcs=data.wcs, unit=data.unit, uncertainty=data.uncertainty)

        if output_filename is not None and i < len(output_filename) and output_filename[i] is not None:
            output_image_task(data, output_filename[i])
        output_data.append(data)
        logger.info("ending level 0.5 core flow")
    return output_data



import os
from datetime import datetime

import astropy
import numpy as np
import pytest
from astropy.io import fits
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS, DistortionLookupTable
from ndcube import NDCube

from punchbowl.data.io import _update_statistics, get_base_file_name, load_ndcube_from_fits, write_ndcube_to_fits
from punchbowl.data.meta import NormalizedMetadata

TESTDATA_DIR = os.path.dirname(__file__)
SAMPLE_FITS_PATH_UNCOMPRESSED = os.path.join(TESTDATA_DIR, "test_data.fits")
SAMPLE_FITS_PATH_COMPRESSED = os.path.join(TESTDATA_DIR, "test_data.fits")
SAMPLE_WRITE_PATH = os.path.join(TESTDATA_DIR, "write_test.fits")
SAMPLE_OMNIBUS_PATH = os.path.join(TESTDATA_DIR, "omniheader.csv")
SAMPLE_LEVEL_PATH = os.path.join(TESTDATA_DIR, "LevelTest.yaml")
SAMPLE_SPACECRAFT_DEF_PATH = os.path.join(TESTDATA_DIR, "spacecraft.yaml")


@pytest.fixture
def sample_ndcube():
    def _sample_ndcube(shape, code="PM1", level="0"):
        data = np.random.random(shape).astype(np.float32)
        sqrt_abs_data = np.sqrt(np.abs(data))
        uncertainty = StdDevUncertainty(np.interp(sqrt_abs_data, (sqrt_abs_data.min(), sqrt_abs_data.max()), (0,1)).astype(np.float32))
        wcs = WCS(naxis=2)
        wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"
        wcs.wcs.cunit = "deg", "deg"
        wcs.wcs.cdelt = 0.1, 0.1
        wcs.wcs.crpix = 0, 0
        wcs.wcs.crval = 1, 1
        wcs.wcs.cname = "HPC lon", "HPC lat"

        meta = NormalizedMetadata.load_template(code, level)
        #meta['DATE-OBS'] = str(datetime(2023, 1, 1, 0, 0, 1))
        meta['DATE-OBS'] = str(datetime(2024, 2, 22, 16, 0, 1))
        #20240222163425
        meta['FILEVRSN'] = "1"
        return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)
    return _sample_ndcube


def test_write_data(sample_ndcube):
    cube = sample_ndcube((50, 50))
    cube.meta["LEVEL"] = "1"
    cube.meta["TYPECODE"] = "CL"
    cube.meta["OBSRVTRY"] = "1"
    cube.meta["PIPEVRSN"] = "0.1"
    cube.meta["DATE-OBS"] = str(datetime.now())
    cube.meta["DATE-END"] = str(datetime.now())

    write_ndcube_to_fits(cube, SAMPLE_WRITE_PATH)
    assert os.path.isfile(SAMPLE_WRITE_PATH)


def test_generate_data_statistics_from_zeros():
    w = WCS(naxis=2)
    m = NormalizedMetadata.load_template("PM1", "0")
    m.history.add_now("Test", "does it write?")
    m.history.add_now("Test", "how about twice?")
    m['DESCRPTN'] = 'This is a test!'
    m['CHECKSUM'] = ''
    m['DATASUM'] = ''
    m.delete_section("World Coordinate System")

    sample_data = NDCube(data=np.zeros((2048, 2048),dtype=np.int16), wcs=w, meta=m)

    _update_statistics(sample_data)

    assert sample_data.meta['DATAZER'].value == 2048*2048

    assert sample_data.meta['DATASAT'].value == 0

    assert sample_data.meta['DATAAVG'].value == -999.0
    assert sample_data.meta['DATAMDN'].value == -999.0
    assert sample_data.meta['DATASIG'].value == -999.0

    percentile_percentages = [1, 10, 25, 50, 75, 90, 95, 98, 99]
    percentile_values = [-999.0 for _ in percentile_percentages]

    for percent, value in zip(percentile_percentages, percentile_values):
        assert sample_data.meta[f'DATAP{percent:02d}'].value == value

    assert sample_data.meta['DATAMIN'].value == 0.0
    assert sample_data.meta['DATAMAX'].value == 0.0


def test_generate_data_statistics(sample_ndcube):
    cube = sample_ndcube((50, 50))
    _update_statistics(cube)

    nonzero_sample_data = cube.data[np.where(cube.data != 0)].flatten()

    assert cube.meta['DATAZER'].value == len(np.where(cube.data == 0)[0])

    assert cube.meta['DATASAT'].value == len(np.where(cube.data >= cube.meta['DSATVAL'].value)[0])

    assert cube.meta['DATAAVG'].value == np.mean(nonzero_sample_data)
    assert cube.meta['DATAMDN'].value == np.median(nonzero_sample_data)
    assert cube.meta['DATASIG'].value == np.std(nonzero_sample_data)

    percentile_percentages = [1, 10, 25, 50, 75, 90, 95, 98, 99]
    percentile_values = np.percentile(nonzero_sample_data, percentile_percentages)

    for percent, value in zip(percentile_percentages, percentile_values):
        assert cube.meta[f'DATAP{percent:02d}'].value == value

    assert cube.meta['DATAMIN'].value == float(cube.data.min())
    assert cube.meta['DATAMAX'].value == float(cube.data.max())


def test_filename_base_generation(sample_ndcube):
    cube = sample_ndcube((50, 50))
    actual = get_base_file_name(cube)
    expected = "PUNCH_L0_PM1_20240222160001_v1"
    assert actual == expected


def test_has_typecode():
    meta = NormalizedMetadata.load_template("CFM", "3")
    meta["DATE-OBS"] = str(datetime.now())
    assert "TYPECODE" in meta


def test_load_punchdata_with_history():
    data = np.ones((10, 10), dtype=np.uint16)
    meta = NormalizedMetadata.load_template("CR4", "0")
    meta['DATE-OBS'] = str(datetime.now())
    meta.history.add_now("test", "this is a test!")
    meta.history.add_now("test", "this is a second test!")
    wcs = WCS({"CRVAL1": 0.0,
                     "CRVAL2": 0.0,
                     "CRPIX1": 2047.5,
                     "CRPIX2": 2047.5,
                     "CDELT1": 0.0225,
                     "CDELT2": 0.0225,
                     "CUNIT1": "deg",
                     "CUNIT2": "deg",
                     "CTYPE1": "HPLN-ARC",
                     "CTYPE2": "HPLT-ARC"})
    obj = NDCube(data=data, wcs=wcs, meta=meta)

    assert "OBSCODE" in obj.meta.fits_keys
    file_path = get_base_file_name(obj) + ".fits"
    write_ndcube_to_fits(obj, file_path, overwrite=True)
    reloaded = load_ndcube_from_fits(file_path)
    assert isinstance(reloaded, NDCube)
    assert len(reloaded.meta.history) == 2
    assert reloaded.data.shape == (10, 10)
    assert np.all(reloaded.data == 1)
    os.remove(file_path)


def make_empty_distortion_model(num_bins: int, image: np.ndarray) -> (DistortionLookupTable, DistortionLookupTable):
    """ Create an empty distortion table

    Parameters
    ----------
    num_bins : int
        number of histogram bins in the distortion model, i.e. the size of the distortion model is (num_bins, num_bins)
    image : np.ndarray
        image to create a distortion model for

    Returns
    -------
    (DistortionLookupTable, DistortionLookupTable)
        x and y distortion models
    """
    # make an initial empty distortion model
    r = np.linspace(0, image.shape[0], num_bins + 1)
    c = np.linspace(0, image.shape[1], num_bins + 1)
    r = (r[1:] + r[:-1]) / 2
    c = (c[1:] + c[:-1]) / 2

    err_px, err_py = r, c
    err_x = np.zeros((num_bins, num_bins))
    err_y = np.zeros((num_bins, num_bins))

    cpdis1 = DistortionLookupTable(
        -err_x.astype(np.float32), (0, 0), (err_px[0], err_py[0]), ((err_px[1] - err_px[0]), (err_py[1] - err_py[0]))
    )
    cpdis2 = DistortionLookupTable(
        -err_y.astype(np.float32), (0, 0), (err_px[0], err_py[0]), ((err_px[1] - err_px[0]), (err_py[1] - err_py[0]))
    )
    return cpdis1, cpdis2


def test_write_punchdata_with_distortion():
    data = np.ones((2048, 2048), dtype=np.uint16)
    meta = NormalizedMetadata.load_template("CR4", "1")
    meta['DATE-OBS'] = str(datetime.now())
    meta.history.add_now("test", "this is a test!")
    meta.history.add_now("test", "this is a second test!")
    wcs = WCS({"CRVAL1": 0.0,
                     "CRVAL2": 0.0,
                     "CRPIX1": 2047.5,
                     "CRPIX2": 2047.5,
                     "CDELT1": 0.0225,
                     "CDELT2": 0.0225,
                     "CUNIT1": "deg",
                     "CUNIT2": "deg",
                     "CTYPE1": "HPLN-ARC",
                     "CTYPE2": "HPLT-ARC"})
    cpdis1, cpdis2 = make_empty_distortion_model(100, data)
    wcs.cpdis1 = cpdis1
    wcs.cpdis2 = cpdis2
    obj = NDCube(data=data, wcs=wcs, meta=meta)
    file_path = get_base_file_name(obj) + ".fits"
    write_ndcube_to_fits(obj, file_path, overwrite=True)

    with fits.open(file_path) as hdul:
        assert len(hdul) == 5

    loaded_cube = load_ndcube_from_fits(file_path)
    assert loaded_cube.wcs.has_distortion



import os
from datetime import datetime

import astropy
import astropy.units as u
import numpy as np
from astropy.coordinates import GCRS, EarthLocation, SkyCoord, get_sun
from astropy.io import fits
from astropy.time import Time
from astropy.wcs import WCS
from ndcube import NDCube
from sunpy.coordinates import frames

from punchbowl.data.meta import NormalizedMetadata
from punchbowl.data.wcs import (
    calculate_celestial_wcs_from_helio,
    calculate_helio_wcs_from_celestial,
    calculate_pc_matrix,
    load_trefoil_wcs,
)

_ROOT = os.path.abspath(os.path.dirname(__file__))


def test_sun_location():
    time_current = Time(datetime.utcnow())

    skycoord_sun = astropy.coordinates.get_sun(time_current)

    skycoord_origin = SkyCoord(0*u.deg, 0*u.deg,
                              frame=frames.Helioprojective,
                              obstime=time_current,
                              observer='earth')

    # with frames.Helioprojective.assume_spherical_screen(skycoord_origin.observer):
    skycoord_origin_celestial = skycoord_origin.transform_to(GCRS)

    # with frames.Helioprojective.assume_spherical_screen(skycoord_origin.observer):
    assert skycoord_origin_celestial.separation(skycoord_sun) < 1 * u.arcsec
    assert skycoord_origin.separation(skycoord_sun) < 1 * u.arcsec


def test_sun_location_gcrs():
    time_current = Time(datetime.utcnow())

    skycoord_sun = astropy.coordinates.get_sun(time_current)

    skycoord_origin = SkyCoord(0*u.deg, 0*u.deg,
                              frame=frames.Helioprojective,
                              obstime=time_current,
                              observer='earth')

    # with frames.Helioprojective.assume_spherical_screen(skycoord_origin.observer):
    skycoord_origin_helio = skycoord_sun.transform_to(frames.Helioprojective)

    # with frames.Helioprojective.assume_spherical_screen(skycoord_origin.observer):
    assert skycoord_origin_helio.separation(skycoord_sun) < 1 * u.arcsec
    assert skycoord_origin.separation(skycoord_sun) < 1 * u.arcsec

def test_extract_crota():
    pc = calculate_pc_matrix(-10*u.deg, [-1, 2])

    wcs_celestial = WCS(naxis=2)
    wcs_celestial.wcs.ctype = ("RA---AZP", "DEC--AZP")
    wcs_celestial.wcs.cunit = ("deg", "deg")
    wcs_celestial.wcs.cdelt = (-1, 2)
    wcs_celestial.wcs.crpix = (2048.5, 2048.5)
    wcs_celestial.wcs.crval = (0, 0)
    wcs_celestial.wcs.pc = pc
    assert np.allclose(extract_crota_from_wcs(wcs_celestial), -10 * u.deg)

def test_extract_crota_helio():
    pc = calculate_pc_matrix(10*u.deg, [0.0225, 0.0225])
    wcs_helio = WCS(naxis=2)
    wcs_helio.wcs.ctype = ("HPLN-AZP", "HPLT-AZP")
    wcs_helio.wcs.cunit = ("deg", "deg")
    wcs_helio.wcs.cdelt = (0.0225, 0.0225)
    wcs_helio.wcs.crpix = (2048.5, 2048.5)
    wcs_helio.wcs.crval = (0, 0)
    wcs_helio.wcs.pc = pc
    assert np.allclose(extract_crota_from_wcs(wcs_helio), 10 * u.deg)

def test_wcs_many_point_2d_check():
    m = NormalizedMetadata.load_template("CTM", "2")
    date_obs = Time("2024-01-01T00:00:00", format='isot', scale='utc')
    m['DATE-OBS'] = str(date_obs)

    sun_radec = get_sun(date_obs)
    wcs_celestial = WCS({"CRVAL1": sun_radec.ra.to(u.deg).value,
                         "CRVAL2": sun_radec.dec.to(u.deg).value,
                         "CRPIX1": 2047.5,
                         "CRPIX2": 2047.5,
                         "CDELT1": -0.0225,
                         "CDELT2": 0.0225,
                         "CUNIT1": "deg",
                         "CUNIT2": "deg",
                         "CTYPE1": "RA---ARC",
                         "CTYPE2": "DEC--ARC"})

    # we're at the center of the Earth so let's try that
    test_loc = EarthLocation.from_geocentric(0, 0, 0, unit=u.m)
    test_gcrs = SkyCoord(test_loc.get_gcrs(date_obs))

    wcs_helio, _ = calculate_helio_wcs_from_celestial(wcs_celestial, date_obs, (4096, 4096))

    npoints = 20
    input_coords = np.stack([
                             np.linspace(0, 4096, npoints).astype(int),
                             np.linspace(0, 4096, npoints).astype(int)], axis=1)

    points_celestial = wcs_celestial.all_pix2world(input_coords, 0)
    points_helio = wcs_helio.all_pix2world(input_coords, 0)

    output_coords = []
    intermediates = []
    for c_pix, c_celestial, c_helio in zip(input_coords, points_celestial, points_helio):
        skycoord_celestial = SkyCoord(c_celestial[0] * u.deg, c_celestial[1] * u.deg,
                                      frame=GCRS,
                                      obstime=date_obs,
                                      observer=test_gcrs,
                                      obsgeoloc=test_gcrs.cartesian,
                                      obsgeovel=test_gcrs.velocity.to_cartesian(),
                                      distance=test_gcrs.hcrs.distance
                                      )

        intermediate = skycoord_celestial.transform_to(frames.Helioprojective)
        intermediates.append(intermediate)
        output_coords.append(wcs_helio.all_world2pix(intermediate.data.lon.to(u.deg).value,
                                                     intermediate.data.lat.to(u.deg).value, 0))

    output_coords = np.array(output_coords)
    print(intermediates[0].Tx.deg,  intermediates[0].Ty.deg, points_helio[0])
    distances = np.linalg.norm(input_coords - output_coords, axis=1)
    assert np.mean(distances) < 0.1


def test_wcs_many_point_3d_check():
    m = NormalizedMetadata.load_template("PSM", "3")
    date_obs = Time("2024-01-01T00:00:00", format='isot', scale='utc')
    m['DATE-OBS'] = str(date_obs)
    sun_radec = get_sun(date_obs)

    wcs_celestial = WCS({"CRVAL1": sun_radec.ra.to(u.deg).value,
                         "CRVAL2": sun_radec.dec.to(u.deg).value,
                         "CRPIX1": 2047.5,
                         "CRPIX2": 2047.5,
                         "CRPIX3": 0,
                         "CDELT1": -0.0225,
                         "CDELT2": 0.0225,
                         "CDELT3": 1.0,
                         "CUNIT1": "deg",
                         "CUNIT2": "deg",
                         "CUNIT3": "",
                         "CTYPE1": "RA---ARC",
                         "CTYPE2": "DEC--ARC",
                         "CTYPE3": "STOKES"})

    # we're at the center of the Earth so let's try that
    test_loc = EarthLocation.from_geocentric(0, 0, 0, unit=u.m)
    test_gcrs = SkyCoord(test_loc.get_gcrs(date_obs))


    wcs_helio, _ = calculate_helio_wcs_from_celestial(wcs_celestial, date_obs,(2, 4096, 4096))

    npoints = 20
    input_coords = np.stack([
                             np.linspace(0, 4096, npoints).astype(int),
                             np.linspace(0, 4096, npoints).astype(int),
                             np.ones(npoints, dtype=int),], axis=1)

    points_celestial = wcs_celestial.all_pix2world(input_coords, 0)
    points_helio = wcs_helio.all_pix2world(input_coords, 0)

    output_coords = []
    for c_pix, c_celestial, c_helio in zip(input_coords, points_celestial, points_helio):
        skycoord_celestial = SkyCoord(c_celestial[0] * u.deg, c_celestial[1] * u.deg,
                                      frame=GCRS,
                                      obstime=date_obs,
                                      observer=test_gcrs,
                                      obsgeoloc=test_gcrs.cartesian,
                                      obsgeovel=test_gcrs.velocity.to_cartesian(),
                                      distance=test_gcrs.hcrs.distance
                                      )

        intermediate = skycoord_celestial.transform_to(frames.Helioprojective)
        output_coords.append(wcs_helio.all_world2pix(intermediate.data.lon.to(u.deg).value,
                                                     intermediate.data.lat.to(u.deg).value, 2, 0))

    output_coords = np.array(output_coords)
    distances = np.linalg.norm(input_coords - output_coords, axis=1)
    assert np.nanmean(distances) < 0.1


def test_load_trefoil_wcs():
    trefoil_wcs, trefoil_shape = load_trefoil_wcs()
    assert trefoil_shape == (4096, 4096)
    assert isinstance(trefoil_wcs, WCS)


def test_helio_celestial_wcs():
    # header = fits.Header.fromtextfile(os.path.join(_ROOT, "example_header.txt"))
    #
    # date_obs = Time(header['DATE-OBS'])
    date_obs = Time("2021-06-20T00:00:00.000", format='isot', scale='utc')
    #
    wcs_helio = WCS({"CRVAL1": 0.0,
                     "CRVAL2": 0.0,
                     "CRPIX1": 2047.5,
                     "CRPIX2": 2047.5,
                     "CRPIX3": 0.0,
                     "CDELT1": 0.0225,
                     "CDELT2": 0.0225,
                     "CDELT3": 1.0,
                     "CUNIT1": "deg",
                     "CUNIT2": "deg",
                     "CTYPE1": "HPLN-ARC",
                     "CTYPE2": "HPLT-ARC",
                     "CTYPE3": "STOKES"})
    # wcs_helio = WCS(header)
    wcs_celestial = calculate_celestial_wcs_from_helio(wcs_helio, date_obs, (3, 4096, 4096))

    # wcs_celestial2 = WCS(header, key='A')

    test_loc = EarthLocation.from_geocentric(0, 0, 0, unit=u.m)
    test_gcrs = SkyCoord(test_loc.get_gcrs(date_obs))

    npoints = 20
    input_coords = np.stack([
                             np.linspace(0, 4096, npoints).astype(int),
                             np.linspace(0, 4096, npoints).astype(int),
                             np.ones(npoints, dtype=int),],
        axis=1)

    points_celestial = wcs_celestial.all_pix2world(input_coords, 0)
    points_helio = wcs_helio.all_pix2world(input_coords, 0)

    output_coords = []
    for c_pix, c_celestial, c_helio in zip(input_coords, points_celestial, points_helio):
        skycoord_celestial = SkyCoord(c_celestial[0] * u.deg, c_celestial[1] * u.deg,
                                      frame=GCRS,
                                      obstime=date_obs,
                                      observer="earth",
                                      # observer=test_gcrs,
                                      # obsgeoloc=test_gcrs.cartesian,
                                      # obsgeovel=test_gcrs.velocity.to_cartesian(),
                                      # distance=test_gcrs.hcrs.distance
                                      )

        intermediate = skycoord_celestial.transform_to(frames.Helioprojective(observer='earth', obstime=date_obs))
        # final = SkyCoord(intermediate.Tx, intermediate.Ty, frame="helioprojective", observer=intermediate.observer, obstime=intermediate.obstime)
        output_coords.append(wcs_helio.all_world2pix(intermediate.Tx.to(u.deg).value,
                                                     intermediate.Ty.to(u.deg).value,
                                                     2,
                                                     0))

    output_coords = np.array(output_coords)
    distances = np.linalg.norm(input_coords - output_coords, axis=1)

    assert np.nanmean(distances) < 0.1

from punchbowl.data.wcs import extract_crota_from_wcs, get_p_angle


def test_back_and_forth_wcs_from_celestial():
    date_obs = Time("2024-01-01T00:00:00", format='isot', scale='utc')
    sun_radec = get_sun(date_obs)
    pc = calculate_pc_matrix(0, [0.0225, 0.0225])
    wcs_celestial = WCS({"CRVAL1": sun_radec.ra.to(u.deg).value,
                         "CRVAL2": sun_radec.dec.to(u.deg).value,
                         "CRPIX1": 2047.5,
                         "CRPIX2": 2047.5,
                         "CDELT1": -0.0225,
                         "CDELT2": 0.0225,
                         "CUNIT1": "deg",
                         "CUNIT2": "deg",
                         "CTYPE1": "RA---AZP",
                         "CTYPE2": "DEC--AZP",
                         "PC1_1": pc[0, 0],
                         "PC1_2": pc[0, 1],
                         "PC2_1": pc[1, 0],
                         "PC2_2": pc[1, 1]
                         })

    print(extract_crota_from_wcs(wcs_celestial))
    wcs_helio, p_angle = calculate_helio_wcs_from_celestial(wcs_celestial, date_obs, (10, 10))
    print(extract_crota_from_wcs(wcs_helio))
    wcs_celestial_recovered = calculate_celestial_wcs_from_helio(wcs_helio, date_obs, (10, 10))
    print(extract_crota_from_wcs(wcs_celestial_recovered))

    assert np.allclose(wcs_celestial.wcs.crval, wcs_celestial_recovered.wcs.crval)
    assert np.allclose(wcs_celestial.wcs.crpix, wcs_celestial_recovered.wcs.crpix)
    assert np.allclose(wcs_celestial.wcs.pc, wcs_celestial_recovered.wcs.pc)
    assert np.allclose(wcs_celestial.wcs.cdelt, wcs_celestial_recovered.wcs.cdelt)


def test_back_and_forth_wcs_from_helio():
    date_obs = Time("2024-01-01T00:00:00", format='isot', scale='utc')

    wcs_helio = WCS({"CRVAL1": 15.0,
                     "CRVAL2": 10.0,
                     "CRPIX1": 2047.5,
                     "CRPIX2": 2047.5,
                     "CDELT1": 0.0225,
                     "CDELT2": 0.0225,
                     "CUNIT1": "deg",
                     "CUNIT2": "deg",
                     "CTYPE1": "HPLN-ARC",
                     "CTYPE2": "HPLT-ARC"})

    print(extract_crota_from_wcs(wcs_helio))
    wcs_celestial = calculate_celestial_wcs_from_helio(wcs_helio.copy(), date_obs, (10, 10))
    print(extract_crota_from_wcs(wcs_celestial))
    wcs_helio_recovered, p_angle = calculate_helio_wcs_from_celestial(wcs_celestial.copy(), date_obs, (10, 10))
    print(extract_crota_from_wcs(wcs_helio_recovered))

    assert np.allclose(wcs_helio.wcs.crval, wcs_helio_recovered.wcs.crval)
    assert np.allclose(wcs_helio.wcs.crpix, wcs_helio_recovered.wcs.crpix)
    assert np.allclose(wcs_helio.wcs.pc, wcs_helio_recovered.wcs.pc)
    assert np.allclose(wcs_helio.wcs.cdelt, wcs_helio_recovered.wcs.cdelt)



import pathlib
from datetime import datetime

import numpy as np
import pytest
from ndcube import NDCube
from prefect.logging import disable_run_logger

from punchbowl.data.tests.test_io import sample_ndcube
from punchbowl.exceptions import LargeTimeDeltaWarning
#from punchbowl.exceptions import InvalidDataError
from punchbowl.level1.stray_light import remove_stray_light_task

THIS_DIRECTORY = pathlib.Path(__file__).parent.resolve()

@pytest.mark.prefect_test()
def test_check_calibration_time_delta_warning(sample_ndcube) -> None:
    """
    If the time between the data of interest and the calibration file is too great, then a warning is raised.
    """

    sample_data = sample_ndcube(shape=(10, 10))
    sample_data.meta['DATE-OBS'].value = str(datetime(2022, 2, 22, 16, 0, 1))
    stray_light_filename = THIS_DIRECTORY / "data" / "PUNCH_L1_SM1_20240222163425_v1.fits"

    with disable_run_logger():
        with pytest.warns(LargeTimeDeltaWarning):
            corrected_punchdata = remove_stray_light_task.fn(sample_data, stray_light_filename)
            assert isinstance(corrected_punchdata, NDCube)


@pytest.mark.prefect_test()
def test_no_straylight_file(sample_ndcube) -> None:
    """
    An invalid vignetting file should be provided. Check that an error is raised.
    """

    sample_data = sample_ndcube(shape=(10, 10), code="CR1", level="0")
    straylight_filename = None

    with disable_run_logger():
        corrected_punchdata = remove_stray_light_task.fn(sample_data, straylight_filename)
        assert isinstance(corrected_punchdata, NDCube)
        assert corrected_punchdata.meta.history[0].comment == 'Stray light correction skipped'



import os

from ndcube import NDCube
from prefect import get_run_logger, task
from regularizepsf.corrector import ArrayCorrector


def correct_psf(
    data: NDCube,
    corrector: ArrayCorrector,
    alpha: float = 2.0,
    epsilon: float = 0.3,
) -> NDCube:
    """Correct PSF."""
    new_data = corrector.correct_image(data.data.copy(), alpha=alpha, epsilon=epsilon)

    data.data[...] = new_data[...]
    # TODO: uncertainty propagation
    return data

@task
def correct_psf_task(
    data_object: NDCube,
    model_path: str | None = None,
) -> NDCube:
    """
    Prefect Task to correct the PSF of an image.

    Parameters
    ----------
    data_object : PUNCHData
        data to operate on
    model_path : str
        path to the PSF model to use in the correction

    Returns
    -------
    PUNCHData
        modified version of the input with the PSF corrected

    """
    logger = get_run_logger()
    logger.info("correct_psf started")

    if model_path is not None:
        corrector = ArrayCorrector.load(model_path)
        data_object = correct_psf(data_object, corrector)
        data_object.meta.history.add_now("LEVEL1-correct_psf",
                                         f"PSF corrected with {os.path.basename(model_path)} model")
        data_object.meta["CALPSF"] = os.path.basename(model_path)
    else:
        data_object.meta.history.add_now("LEVEL1-correct_psf", "Empty model path so no correction applied")
        logger.info("No model path so PSF correction is skipped")

    logger.info("correct_psf finished")
    return data_object



import astropy.constants as const
import astropy.units as u
import numpy as np
from astropy.wcs import WCS
from numpy import ndarray

MSB = u.def_unit("MSB", 2.0090000E7 * u.W / u.m ** 2 / u.sr)


def calculate_image_pixel_area(wcs: WCS, data_shape: tuple[int, int]) -> u.sr:
    """Calculate the sky area of every pixel in an image according to its WCS."""
    xx, yy = np.meshgrid(np.arange(data_shape[0]+1), np.arange(data_shape[1]+1))
    coords = wcs.pixel_to_world(xx, yy)
    tx = coords.Tx.to(u.deg).value
    ty = coords.Ty.to(u.deg).value
    pixel_area = abs(tx[:, 1:] - tx[:, :-1])[1:, :] * abs(ty[1:] - ty[:-1])[:, 1:]
    return pixel_area * u.degree * u.degree


def msb_to_dn(data: ndarray,
              data_wcs: WCS,
              gain: float = 4.9 * u.photon / u.DN,
              wavelength: float = 530. * u.nm,
              exposure: float = 49 * u.s,
              aperture: float = 49.57 * u.mm**2,
              ) -> ndarray:
    """Convert mean solar brightness to DNs."""
    energy_per_photon = (const.h * const.c / wavelength).to(u.J) / u.photon
    photon_flux = MSB / energy_per_photon
    pixel_scale = calculate_image_pixel_area(data_wcs, data.shape).to(u.sr) / u.pixel
    photon_count = (photon_flux * exposure * aperture * pixel_scale * u.pixel).decompose()
    return (data * photon_count / gain).astype(int)


def dn_to_msb(data: ndarray,
              data_wcs: WCS,
              gain: float = 4.9 * u.photon / u.DN,
              wavelength: float = 530. * u.nm,
              exposure: float = 49 * u.s,
              aperture: float = 34 * u.mm**2,
              ) -> ndarray:
    """Convert DN to mean solar brightness."""
    energy_per_photon = (const.h * const.c / wavelength).to(u.J) / u.photon
    photon_flux = MSB / energy_per_photon
    pixel_scale = calculate_image_pixel_area(data_wcs, data.shape).to(u.sr) / u.pixel
    photon_count = (photon_flux * exposure * aperture * pixel_scale * u.pixel).decompose()
    return data * gain / photon_count



from punchbowl import __version__

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "punchbowl"
copyright = "2024, PUNCH Science Operations Center"
author = "PUNCH Science Operations Center"

# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

extensions = ["autoapi.extension",
              "sphinx.ext.autodoc",
              "sphinx.ext.napoleon",
              "sphinx_favicon",
              "nbsphinx",
              "sphinxcontrib.mermaid",
              "IPython.sphinxext.ipython_console_highlighting"]

nbsphinx_requirejs_path = ""

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
html_show_sourcelink = False
html_static_path = ["_static"]
html_theme_options = {
    "use_edit_page_button": True,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/punch-mission/punchbowl",
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        }
    ],
    "show_nav_level": 1,
    "show_toc_level": 3,
    "logo": {
        "text": "punchbowl",
        "image_light": "_static/logo.png",
        "image_dark": "_static/logo.png",
    }
}
html_context = {
    # "github_url": "https://github.com", # or your GitHub Enterprise site
    "github_user": "punch-mission",
    "github_repo": "punchbowl",
    "github_version": "main",
    "doc_path": "docs/source/",
}


autoapi_dirs = ["../../punchbowl"]

favicons = ["favicon.ico"]

mermaid_verbose = True




from ndcube import NDCube
from prefect import flow, get_run_logger

from punchbowl.level3.f_corona_model import subtract_f_corona_background_task
from punchbowl.level3.polarization import convert_polarization
from punchbowl.level3.stellar import subtract_starfield_background_task
from punchbowl.util import load_image_task, output_image_task


@flow(validate_parameters=False)
def level3_core_flow(data_list: list[str] | list[NDCube],
                     f_corona_model_path: str | None,
                     starfield_background_path: str | None,
                     output_filename: str | None = None) -> list[NDCube]:
    """Level 3 core flow."""
    logger = get_run_logger()

    logger.info("beginning level 3 core flow")
    data_list = [load_image_task(d) if isinstance(d, str) else d for d in data_list]
    data_list = [subtract_f_corona_background_task(d, f_corona_model_path) for d in data_list]
    data_list = [subtract_starfield_background_task(d, starfield_background_path) for d in data_list]
    data_list = [convert_polarization(d) for d in data_list]
    # TODO: build low noise products
    logger.info("ending level 3 core flow")

    if output_filename is not None:
        output_image_task(data_list[0], output_filename)

    return data_list



import warnings
from datetime import datetime

import numpy as np
from ndcube import NDCube
from prefect import get_run_logger, task

from punchbowl.data import load_ndcube_from_fits
from punchbowl.exceptions import InvalidDataError


@task
def query_f_corona_model_source(
    polarizer: str, product: str, start_datetime: datetime, end_datetime: datetime,
) -> list[str]:
    """
    Create a list of files for later F corona model generation.

    Creates a list of files based between a start date/time (start_datetime)
    and an end date/time (end_datetime) for a specified polarizer and
    PUNCH_product. The start and end times can both be input explicitly,
    individually, or derived from a mid time.

    if start_datetime (datetime object) and an end_datetime(datetime object)
    are specified a list of files between those dates is produced for the
    specified polarizer and PUNCH_product.


    Parameters
    ----------
    polarizer : string [= 'clear', '-60', '60', '0' ]
        input a string specifying the type of polarizer to search for

    product : string [= 'mosaic', 'nfi']
        product code to analyze

    start_datetime : datetime
        input a start_datetime of interest.

    end_datetime : datetime
        input a start_datetime of interest.

    Returns
    -------
    file_list : [list]
        returns a list of files over the specified period for the specified
        polarizer.

    """
    logger = get_run_logger()
    logger.info("query_f_corona_model_source started")

    # Check for polarizer input, and if valid
    if polarizer not in ["clear", "-60", "60", "0"]:
        msg = f"input polarizer expected as string: 'clear', '-60', '60', '0'. Found {polarizer}."
        raise ValueError(msg)

    # Check for PUNCH input, and if valid
    if product not in ["nfi", "mosaic"]:
        msg = f"input PUNCH_product expected as string:  'nfi', 'mosaic'. Found {product}"
        raise ValueError(msg)

    file_list = []
    # file_list=session.query(File).where(and_(File.state == "created",
    #                                           File.date_obs >= date.today() + relativedelta(months=-1))).all()

    # Check if files found in range
    if len(file_list) == 0:
        warnings.warn(f"No files found over specified dates: from: {start_datetime} to: {end_datetime}.", stacklevel=2)

    logger.info("query_f_corona_model_source finished")

    return file_list


def _zvalue_from_index(arr, ind):  # noqa: ANN202, ANN001
    """
    Do math.

    Private helper function to work around the limitation of np.choose() by employing np.take().
    arr has to be a 3D array
    ind has to be a 2D array containing values for z-indicies to take from arr
    See: http://stackoverflow.com/a/32091712/4169585
    This is faster and more memory efficient than using the ogrid based solution with fancy indexing.
    """
    # get number of columns and rows
    _, n_cols, n_rows = arr.shape

    # get linear indices and extract elements with np.take()
    idx = n_cols*n_rows*ind + n_rows*np.arange(n_rows)[:,None] + np.arange(n_cols)
    return np.take(arr, idx)


def nan_percentile(arr: np.ndarray, q: list[float] | float) -> np.ndarray:
    """Calculate the nan percentile faster of a 3D cube."""
    # np.nanpercentile is slow so use this: https://krstn.eu/np.nanpercentile()-there-has-to-be-a-faster-way/

    # valid (non NaN) observations along the first axis
    valid_obs = np.sum(np.isfinite(arr), axis=0)
    # replace NaN with maximum
    max_val = np.nanmax(arr)
    arr[np.isnan(arr)] = max_val
    # sort - former NaNs will move to the end
    arr = np.sort(arr, axis=0)

    # loop over requested quantiles
    if isinstance(q, list):
        qs = []
        qs.extend(q)
    else:
        qs = [q]
    if len(qs) < 2:
        quant_arr = np.zeros(shape=(arr.shape[1], arr.shape[2]))
    else:
        quant_arr = np.zeros(shape=(len(qs), arr.shape[1], arr.shape[2]))

    result = []
    for i in range(len(qs)):
        quant = qs[i]
        # desired position as well as floor and ceiling of it
        k_arr = (valid_obs - 1) * (quant / 100.0)
        f_arr = np.floor(k_arr).astype(np.int32)
        c_arr = np.ceil(k_arr).astype(np.int32)
        fc_equal_k_mask = f_arr == c_arr

        # linear interpolation (like numpy percentile) takes the fractional part of desired position
        floor_val = _zvalue_from_index(arr=arr, ind=f_arr) * (c_arr - k_arr)
        ceil_val = _zvalue_from_index(arr=arr, ind=c_arr) * (k_arr - f_arr)

        quant_arr = floor_val + ceil_val
        # if floor == ceiling take floor value
        quant_arr[fc_equal_k_mask] = _zvalue_from_index(arr=arr, ind=k_arr.astype(np.int32))[fc_equal_k_mask]

        result.append(quant_arr)

    return result


@task
def construct_f_corona_background(
    data_list: list[str],
    layer: int,
) -> NDCube:
    """Build f corona background model."""
    logger = get_run_logger()
    logger.info("construct_f_corona_background started")

    # create an empty array to fill with data
    #   open the first file in the list to ge the shape of the file
    if len(data_list) == 0:
        msg = "data_list cannot be empty"
        raise ValueError(msg)

    output = load_ndcube_from_fits(data_list[0])
    output_wcs = output.wcs
    output_meta = output.meta
    output_mask = output.mask

    data_shape = output.data[layer].shape

    number_of_data_frames = len(data_list)
    data_cube = np.empty((number_of_data_frames, *data_shape), dtype=float)
    uncertainty_cube = np.empty((number_of_data_frames, *data_shape), dtype=float)
    meta_list = []

    for i, address_out in enumerate(data_list):
        data_object = load_ndcube_from_fits(address_out)
        data_cube[i, :, :] = data_object.data[layer]
        uncertainty_cube[i, :, :] = data_object.uncertainty.array[layer]
        meta_list.append(data_object.meta)


    # create an output PUNCHdata object
    f_background = np.zeros_like(data_cube[0])
    output = NDCube(f_background, wcs=output_wcs, meta=output_meta, mask=output_mask)

    logger.info("construct_f_corona_background finished")
    output.meta.history.add_now("LEVEL3-construct_f_corona_background", "constructed f corona model")

    return output


def subtract_f_corona_background(data_object: NDCube,
                                 before_f_background_model: NDCube,
                                 after_f_background_model: NDCube ) -> NDCube:
    """Subtract f corona background."""
    # check dimensions match
    if data_object.data.shape != before_f_background_model.data.shape:
        msg = (
            "f_background_subtraction expects the data_object and"
            "f_background arrays to have the same dimensions."
            f"data_array dims: {data_object.data.shape} "
            f"and before_f_background_model dims: {before_f_background_model.data.shape}"
        )
        raise InvalidDataError(
            msg,
        )

    if data_object.data.shape != after_f_background_model.data.shape:
        msg = (
            "f_background_subtraction expects the data_object and"
            "f_background arrays to have the same dimensions."
            f"data_array dims: {data_object.data.shape} "
            f"and after_f_background_model dims: {after_f_background_model.data.shape}"
        )
        raise InvalidDataError(
            msg,
        )

    before_date = before_f_background_model.meta.datetime.timestamp()
    after_date = after_f_background_model.meta.datetime.timestamp()
    observation_date = data_object.meta.datetime.timestamp()

    if before_date > observation_date:
        msg = "Before F corona model was after the observation date"
        raise InvalidDataError(msg)

    if after_date < observation_date:
        msg = "After F corona model was before the observation date"
        raise InvalidDataError(msg)

    interpolated_model = ((after_f_background_model.data - before_f_background_model.data)
                          * (observation_date - before_date) / (after_date - before_date)
                          + before_f_background_model.data)
    interpolated_model[np.isinf(data_object.uncertainty.array)] = 0

    bkg_subtracted_data = data_object.data - interpolated_model

    data_object.data[...] = bkg_subtracted_data
    data_object.uncertainty.array[:, :] -= interpolated_model
    return data_object

@task
def subtract_f_corona_background_task(observation: NDCube,
                                      before_f_background_model_path: str,
                                      after_f_background_model_path: str) -> NDCube:
    """
    Subtracts a background f corona model from an observation.

    This algorithm linearly interpolates between the before and after models.

    Parameters
    ----------
    observation : NDCube
        an observation to subtract an f corona model from

    before_f_background_model_path : str
        path to a NDCube f corona background map before the observation

    after_f_background_model_path : str
        path to a NDCube f corona background map after the observation

    Returns
    -------
    NDCube
        A background subtracted data frame

    """
    logger = get_run_logger()
    logger.info("subtract_f_corona_background started")


    before_f_corona_model = load_ndcube_from_fits(before_f_background_model_path)
    after_f_corona_model = load_ndcube_from_fits(after_f_background_model_path)

    output = subtract_f_corona_background(observation, before_f_corona_model, after_f_corona_model)
    output.meta.history.add_now("LEVEL3-subtract_f_corona_background", "subtracted f corona background")

    logger.info("subtract_f_corona_background finished")

    return output


def create_empty_f_background_model(data_object: NDCube) -> np.ndarray:
    """Create an empty background model."""
    return np.zeros_like(data_object.data)




import numpy as np
import remove_starfield
from ndcube import NDCube
from prefect import get_run_logger, task
from remove_starfield import Starfield
from remove_starfield.reducers import GaussianReducer

from punchbowl.data import NormalizedMetadata, load_ndcube_from_fits


def generate_starfield_background(
        data_object: NDCube,
        n_sigma: float = 5,
        map_scale: float = 0.01,
        target_mem_usage: float = 1000) -> NDCube:
    """Create a background starfield_bg map from a series of PUNCH images over a long period of time."""
    logger = get_run_logger()
    logger.info("construct_starfield_background started")

    # create an empty array to fill with data
    #   open the first file in the list to ge the shape of the file
    if len(data_object.data) == 0:
        msg = "data_list cannot be empty"
        raise ValueError(msg)

    starfield_bg = remove_starfield.build_starfield_estimate(
        data_object,
        attribution=True,
        frame_count=True,
        reducer=GaussianReducer(n_sigma=n_sigma),
        map_scale=map_scale,
        processor=remove_starfield.ImageProcessor(),
        target_mem_usage=target_mem_usage)

    # create an output PUNCHdata object
    meta_norm = NormalizedMetadata.from_fits_header(starfield_bg["meta"])
    output = NDCube(starfield_bg.starfield, wcs=starfield_bg.wcs, meta=meta_norm)

    logger.info("construct_starfield_background finished")
    output.meta.history.add_now("LEVEL3-starfield_background", "constructed starfield_bg model")

    return output


def subtract_starfield_background(data_object: NDCube, starfield_background_model: Starfield) -> NDCube:
    """Subtract starfield background."""
    starfield_subtracted_data = Starfield.subtract_from_image(starfield_background_model,
                                                              data_object,
                                                              processor=remove_starfield.ImageProcessor())

    starfield_subtracted_uncertainty = Starfield.subtract_from_image(starfield_background_model,
                                                              NDCube(data=data_object.uncertainty.array,
                                                                     wcs=data_object.wcs,
                                                                     meta=data_object.meta),
                                                              processor=remove_starfield.ImageProcessor())

    data_object.data[...] = starfield_subtracted_data.subtracted
    data_object.uncertainty.array[...] = starfield_subtracted_uncertainty.subtracted

    return data_object


@task
def subtract_starfield_background_task(data_object: NDCube,
                                       starfield_background_path: str | None) -> NDCube:
    """
    Subtracts a background starfield from an input data frame.

    checks the dimensions of input data frame and background starfield match and
    subtracts the background starfield from the data frame of interest.

    Parameters
    ----------
    data_object : punchbowl.data.PUNCHData
        A PUNCHData data frame to be background subtracted

    starfield_background_path : str
        path to a PUNCHData background starfield map

    Returns
    -------
    'punchbowl.data.PUNCHData'
        A background starfield subtracted data frame

    """
    logger = get_run_logger()
    logger.info("subtract_starfield_background started")

    if starfield_background_path is None:
        output = data_object
        output.meta.history.add_now("LEVEL3-fcorona-subtraction",
                                           "F corona subtraction skipped since path is empty")
    else:
        star_data_array = load_ndcube_from_fits(starfield_background_path).data
        output = subtract_starfield_background(data_object, star_data_array)
        output.meta.history.add_now("LEVEL3-subtract_starfield_background", "subtracted starfield background")

    logger.info("subtract_f_corona_background finished")

    return output


def create_empty_starfield_background(data_object: NDCube) -> np.ndarray:
    """Create an empty starfield background map."""
    return np.zeros_like(data_object.data)



from __future__ import annotations

import os.path

import astropy.units as u
import matplotlib as mpl
import numpy as np
from astropy.io import fits
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from ndcube import NDCube

from punchbowl.data.meta import NormalizedMetadata

_ROOT = os.path.abspath(os.path.dirname(__file__))


def get_base_file_name(cube: NDCube) -> str:
    """Determine the base file name without file type extension."""
    obscode = cube.meta["OBSCODE"].value
    file_level = cube.meta["LEVEL"].value
    type_code = cube.meta["TYPECODE"].value
    date_string = cube.meta.datetime.strftime("%Y%m%d%H%M%S")
    file_version = cube.meta["FILEVRSN"].value
    file_version = "1" if file_version == "" else file_version  # file version should never be empty!
    return "PUNCH_L" + file_level + "_" + type_code + obscode + "_" + date_string + "_v" + file_version


def write_ndcube_to_fits(cube: NDCube,
                         filename: str,
                         overwrite: bool = True) -> None:
    """Write an NDCube as a FITS file."""
    if filename.endswith(".fits"):
        _write_fits(cube, filename, overwrite=overwrite)
    else:
        msg = (
            "Filename must have a valid file extension `.fits`"
            f"Found: {os.path.splitext(filename)[1]}"
        )
        raise ValueError(
            msg,
        )


def _write_fits(cube: NDCube, filename: str, overwrite: bool = True, uncertainty_quantize_level: float = 16.0) -> None:
    _update_statistics(cube)

    full_header = cube.meta.to_fits_header(wcs=cube.wcs)

    hdu_data = fits.CompImageHDU(data=cube.data,
                                 header=full_header,
                                 name="Primary data array")
    hdu_uncertainty = fits.CompImageHDU(data=_pack_uncertainty(cube),
                                        header=full_header,
                                        name="Uncertainty array",
                                        quantize_level=uncertainty_quantize_level)

    hdul = cube.wcs.to_fits()
    hdul[0] = fits.PrimaryHDU()
    hdul.insert(1, hdu_data)
    hdul.insert(2, hdu_uncertainty)
    hdul.writeto(filename, overwrite=overwrite, checksum=True)
    hdul.close()

def _pack_uncertainty(cube: NDCube) -> np.ndarray:
    """Compress the uncertainty for writing to file."""
    return np.zeros_like(cube.data) - 999 if cube.uncertainty is None else 1 / (cube.uncertainty.array / cube.data)


def _unpack_uncertainty(uncertainty_array: np.ndarray, data_array: np.ndarray) -> np.ndarray:
    """Uncompress the uncertainty when reading from a file."""
    return (1/uncertainty_array) * data_array


def _write_ql(cube: NDCube, filename: str, overwrite: bool = True) -> None:
    if os.path.isfile(filename) and not overwrite:
        msg = f"File {filename} already exists. If you mean to replace it then use the argument 'overwrite=True'."
        raise OSError(
            msg,
        )

    if cube.data.ndim != 2:
        msg = "Specified output data should have two-dimensions."
        raise ValueError(msg)

    # Scale data array to 8-bit values
    output_data = int(np.fix(np.interp(cube.data, (cube.data.min(), cube.data.max()), (0, 2**8 - 1))))

    # Write image to file
    mpl.image.saveim(filename, output_data)


def _update_statistics(cube: NDCube) -> None:
    """Update image statistics in metadata before writing to file."""
    # TODO - Determine DSATVAL omniheader value in calibrated units for L1+

    cube.meta["DATAZER"] = len(np.where(cube.data == 0)[0])

    cube.meta["DATASAT"] = len(np.where(cube.data >= cube.meta["DSATVAL"].value)[0])

    cube.data[np.logical_not(np.isfinite(cube.data))] = 0
    nonzero_data = cube.data[np.isfinite(cube.data) * (cube.data != 0)].flatten()

    if len(nonzero_data) > 0:
        cube.meta["DATAAVG"] = np.nanmean(nonzero_data).item()
        cube.meta["DATAMDN"] = np.nanmedian(nonzero_data).item()
        cube.meta["DATASIG"] = np.nanstd(nonzero_data).item()
    else:
        cube.meta["DATAAVG"] = -999.0
        cube.meta["DATAMDN"] = -999.0
        cube.meta["DATASIG"] = -999.0

    percentile_percentages = [1, 10, 25, 50, 75, 90, 95, 98, 99]
    percentile_values = np.nanpercentile(nonzero_data, percentile_percentages)
    if np.any(np.isnan(percentile_values)):  # report nan if any of the values are nan
        percentile_values = [-999.0 for _ in percentile_percentages]

    for percent, value in zip(percentile_percentages, percentile_values, strict=False):
        cube.meta[f"DATAP{percent:02d}"] = value

    cube.meta["DATAMIN"] = float(np.nanmin(cube.data))
    cube.meta["DATAMAX"] = float(np.nanmax(cube.data))


def load_ndcube_from_fits(path: str, key: str = " ") -> NDCube:
    """Load an NDCube from a FITS file."""
    with fits.open(path) as hdul:
        primary_hdu = hdul[1]
        data = primary_hdu.data
        header = primary_hdu.header
        # Reset checksum and datasum to match astropy.io.fits behavior
        header["CHECKSUM"] = ""
        header["DATASUM"] = ""
        meta = NormalizedMetadata.from_fits_header(header)
        wcs = WCS(header, hdul, key=key)
        unit = u.ct

        secondary_hdu = hdul[2]
        uncertainty = _unpack_uncertainty(secondary_hdu.data.astype(float), data)
        uncertainty = StdDevUncertainty(uncertainty)

    return NDCube(
        data.view(dtype=data.dtype.newbyteorder()).byteswap().astype(float),
        wcs=wcs,
        uncertainty=uncertainty,
        meta=meta,
        unit=unit,
    )



import os

import numpy as np
from ndcube import NDCube
from prefect import task

from punchbowl.data import load_ndcube_from_fits, write_ndcube_to_fits


def validate_image_is_square(image: np.ndarray) -> None:
    """Check that the input array is square."""
    if not isinstance(image, np.ndarray):
        msg = f"Image must be of type np.ndarray. Found: {type(image)}."
        raise TypeError(msg)
    if len(image.shape) != 2:
        msg = f"Image must be a 2-D array. Input has {len(image.shape)} dimensions."
        raise ValueError(msg)
    if not np.equal(*image.shape):
        msg = f"Image must be square. Found: {image.shape}."
        raise ValueError(msg)


@task
def output_image_task(data: NDCube, output_filename: str) -> None:
    """
    Prefect task to write an image to disk.

    Parameters
    ----------
    data : PUNCHData
        data that is to be written
    output_filename : str
        where to write the file out

    Returns
    -------
    None

    """
    output_dir = os.path.dirname(output_filename)
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)
    write_ndcube_to_fits(data, output_filename)


@task
def load_image_task(input_filename: str) -> NDCube:
    """
    Prefect task to load data for processing.

    Parameters
    ----------
    input_filename : str
        path to file to load

    Returns
    -------
    PUNCHData
        loaded version of the image

    """
    return load_ndcube_from_fits(input_filename)



import numpy as np
import pytest
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from ndcube import NDCube
from remove_starfield import Starfield

from punchbowl.data import NormalizedMetadata
from punchbowl.level3.stellar import (
    generate_starfield_background,
    subtract_starfield_background,
    subtract_starfield_background_task,
)


@pytest.fixture()
def one_data(shape: tuple = (128, 128)) -> NDCube:
    """
    Generate some random data for testing
    """
    data = np.ones(shape)

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 64, 64
    wcs.wcs.crval = 0, 24.75

    meta = NormalizedMetadata(
        {"TYPECODE": "CS", "OBSCODE": "M", "LEVEL": "3", "OBSRVTRY": "0", "DATE-OBS": "2024-04-08T18:40:00"})
    return NDCube(data=data, wcs=wcs, meta=meta, uncertainty=StdDevUncertainty(np.zeros_like(data)))


@pytest.fixture()
def zero_starfield_data(shape: tuple = (256, 256)) -> Starfield:
    """
    Generate some random data for testing
    """
    starfield = np.zeros(shape)

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 128, 128
    wcs.wcs.crval = 0, 24.75

    return Starfield(starfield=starfield, wcs=wcs)


def test_basic_subtraction(one_data: NDCube, zero_starfield_data: Starfield) -> None:
    """

    """
    subtraction_starfield = subtract_starfield_background(one_data, zero_starfield_data)
    subtraction_punchdata = NDCube(data=subtraction_starfield.data, wcs=subtraction_starfield.wcs, meta=subtraction_starfield.meta)
    assert isinstance(subtraction_punchdata, NDCube)
    assert np.allclose(subtraction_punchdata.data, 1)



import os

import numpy as np
from ndcube import NDCube
from prefect import get_run_logger, task

from punchbowl.data import load_ndcube_from_fits


def create_coefficient_image(flat_coefficients: np.ndarray, image_shape: tuple) -> np.ndarray:
    """
    Given a set of coefficients that should apply for every pixel, convert them to the required format.

    Parameters
    ----------
    flat_coefficients : np.ndarray
        A one-dimensional list of coefficients that should apply to every pixel in the image.
        Coefficients should be ordered from the highest power to lowest as expected in `photometric_calibration`, e.g.
        f(i,j) = a+b*IMG[i,j]+c*IMG[i,j]^2 would have flat_coefficients of [c, b, a]
    image_shape : tuple
        A tuple of the shape of the image that will be calibrated using `photometric_calibration`

    Returns
    -------
    np.ndarray
        An image of coefficients that apply to every pixel as expected by `photometric_calibration`

    """
    return np.stack([np.ones(image_shape) * coeff for coeff in flat_coefficients], axis=2)


def create_constant_quartic_coefficients(img_shape: tuple) -> np.ndarray:
    """
    Create a constant coefficients image that preserves the original values, b = 1 and all other coefficients are 0.

    Parameters
    ----------
    img_shape : tuple[Int]
        size of the image to create the coefficients for

    Returns
    -------
    np.ndarray
        An image of coefficients that apply to every pixel as expected by `photometric_calibration`

    """
    return create_coefficient_image(np.array([0, 0, 0, 1, 0]), img_shape)


def photometric_calibration(image: np.ndarray, coefficient_image: np.ndarray) -> np.ndarray:
    """
    Compute a non-linear photometric calibration of PUNCH images.

    Parameters
    ----------
    image : np.ndarray
        Image to be corrected.

    coefficient_image : np.ndarray
        Frame containing uncertainty values.
        The first two dimensions are the spatial dimensions of the image.
        The last dimension iterates over the powers of the coefficients, starting with index 0 being the highest power
        and counting down.

    Returns
    -------
    np.ndarray
        a photometrically corrected frame

    Notes
    -----
    Each instrument is subject to an independent non-linear photometric response,
    which needs to be corrected. The module converts from raw camera digitizer
    number (DN) to photometric units at each pixel. Each pixel is replaced with
    the corresponding value of the quartic polynomial in the current calibration file data
    product for that particular camera.

    A quartic polynomial is applied as follows:

    .. math:: X_{i,j} = a_{i,j}+b_{i,j}*DN_{i,j}+c_{i,j}*DN_{i,j}^2+d_{i,j}*DN_{i,j}^3+e_{i,j}*DN_{i,j}^4

    for each pixel in the detector. Each quantity (a, b, c, d, e) is a function
    of pixel location (i,j), and is generated using dark current and Stim lamp
    maps. a = offset (dark and the bias). b, c, d, e = higher order terms.
    Specifically ``coefficient_image[i,j,:] = [e, d, c, b, a]`` (highest order terms first)

    As each pixel is independent, a quartic fit calibration file of
    dimensions 2k*2k*5 is constructed, with each layer containing one of the five
    polynomial coefficients for each pixel.

    Examples
    --------
    >>> punch_image = np.ones((100,100))
    >>> coefficient_image = create_coefficient_image(np.array([0, 0, 0, 1, 0]), punch_image.shape)
    >>> data = photometric_calibration(punch_image, coefficient_image)

    """
    # inspect dimensions
    if len(image.shape) != 2:
        msg = "`image` must be a 2-D image"
        raise ValueError(msg)

    if len(coefficient_image.shape) != 3:
        msg = "`coefficient_image` must be a 3-D image"
        raise ValueError(msg)

    if coefficient_image.shape[:-1] != image.shape:
        msg = "`coefficient_image` and `image` must have the same shape`"
        raise ValueError(msg)

    # find the number of quartic fit coefficients
    num_coefficients = coefficient_image.shape[2]
    return np.sum(
        [coefficient_image[..., i] * np.power(image, num_coefficients - i - 1) for i in range(num_coefficients)],
        axis=0,
    )


@task
def perform_quartic_fit_task(data_object: NDCube, quartic_coefficients_path: str | None = None) -> NDCube:
    """
    Prefect task to perform the quartic fit calibration on the data.

    Parameters
    ----------
    data_object : PUNCHData
        a data object that needs calibration
    quartic_coefficients_path: Optional[str]
        path to a  cube of coefficients as produced by `create_coefficients_image` or `create_ones_coefficients_image`,
        skips correction if it is None

    Returns
    -------
    PUNCHData
        modified version of the input with the quartic fit correction applied

    See Also
    --------
    photometric_calibration

    """
    logger = get_run_logger()
    logger.info("perform_quartic_fit started")

    if quartic_coefficients_path is not None:
        quartic_coefficients = load_ndcube_from_fits(quartic_coefficients_path)
        new_data = photometric_calibration(data_object.data, quartic_coefficients.data)
        data_object.data[...] = new_data[...]

        new_uncertainty = photometric_calibration(data_object.uncertainty.array, quartic_coefficients.data)
        data_object.uncertainty.array[...] = new_uncertainty

        data_object.meta.history.add_now(
            "LEVEL1-quartic_fit",
            f"Quartic fit correction completed with {os.path.basename(quartic_coefficients_path)}",
        )
        data_object.meta["CALCF"] = os.path.basename(quartic_coefficients_path)
    else:
        data_object.meta.history.add_now("LEVEL1-quartic_fit", "Quartic fit correction skipped since path is empty")

    logger.info("perform_quartic_fit finished")

    return data_object



import pytest
from ndcube import NDCube
from prefect.logging import disable_run_logger
from prefect.testing.utilities import prefect_test_harness

from punchbowl.data.tests.test_io import sample_ndcube
from punchbowl.level2.polarization import resolve_polarization, resolve_polarization_task


@pytest.fixture
def sample_data_triplet(sample_ndcube):
    """
    Generate a list of sample PUNCHData objects for testing polarization resolving
    """

    sample_pd1 = sample_ndcube(shape=(50, 50))
    sample_pd2 = sample_ndcube(shape=(50, 50))
    sample_pd3 = sample_ndcube(shape=(50, 50))
    return [sample_pd1, sample_pd2, sample_pd3]


def test_resolve_polarization(sample_data_triplet):
    """
        Test the resolve polarization function directly
    """

    output_punchdata_list = resolve_polarization(sample_data_triplet)

    assert all(isinstance(output_punchdata, NDCube) for output_punchdata in output_punchdata_list)


@pytest.mark.prefect_test
def test_resolve_polarization_task(sample_data_triplet):
    """
    Test the resolve polarization prefect flow using a test harness
    """

    with disable_run_logger():
        output_punchdata_list = resolve_polarization_task.fn(sample_data_triplet)

    assert all(isinstance(output_punchdata, NDCube) for output_punchdata in output_punchdata_list)



import os
import pathlib
import warnings

from astropy.time import Time
from ndcube import NDCube
from prefect import get_run_logger, task

from punchbowl.data import load_ndcube_from_fits
from punchbowl.exceptions import (
    IncorrectPolarizationStateWarning,
    IncorrectTelescopeWarning,
    InvalidDataError,
    LargeTimeDeltaWarning,
)


@task
def remove_stray_light_task(data_object: NDCube, stray_light_path: pathlib) -> NDCube:
    """
    Prefect task to remove stray light from an image.

    Stray light is light in an optical system which was not intended in the
    design.

    The PUNCH instrument stray light will be mapped periodically as part of the
    ongoing in-flight calibration effort. The stray light maps will be
    generated directly from the L0 and L1 science data. Separating instrumental
    stray light from the F-corona. This has been demonstrated with SOHO/LASCO
    and with STEREO/COR2 observations. It requires an instrumental roll to hold
    the stray light pattern fixed while the F-corona rotates in the field of
    view. PUNCH orbital rolls will be used to create similar effects.

    Uncertainty across the image plane is calculated using a known stray light
    model and the difference between the calculated stray light and the ground
    truth. The uncertainty is convolved with the input uncertainty layer to
    produce the output uncertainty layer.


    Parameters
    ----------
    data_object : PUNCHData
        data to operate on

    stray_light_path: pathlib
        path to stray light model to apply to data

    Returns
    -------
    PUNCHData
        modified version of the input with the stray light removed

    """
    logger = get_run_logger()
    logger.info("remove_stray_light started")

    if stray_light_path is None:
        data_object.meta.history.add_now("LEVEL1-remove_stray_light", "Stray light correction skipped")
    elif not stray_light_path.exists():
        msg = f"File {stray_light_path} does not exist."
        raise InvalidDataError(msg)
    else:
        stray_light_model = load_ndcube_from_fits(stray_light_path)

        stray_light_model_date = Time(stray_light_model.meta["DATE-OBS"].value)
        observation_date = Time(data_object.meta["DATE-OBS"].value)
        if abs((stray_light_model_date - observation_date).to("day").value) > 14:
            msg=f"Calibration file {stray_light_path} contains data created greater than 2 weeks from the obsveration"
            warnings.warn(msg,LargeTimeDeltaWarning)

        if stray_light_model.meta["TELESCOP"].value != data_object.meta["TELESCOP"].value:
            msg=f"Incorrect TELESCOP value within {stray_light_path}"
            warnings.warn(msg, IncorrectTelescopeWarning)
        elif stray_light_model.meta["OBSLAYR1"].value != data_object.meta["OBSLAYR1"].value:
            msg=f"Incorrect polarization state within {stray_light_path}"
            warnings.warn(msg, IncorrectPolarizationStateWarning)
        elif stray_light_model.data.shape != data_object.data.shape:
            msg = f"Incorrect stray light function shape within {stray_light_path}"
            raise InvalidDataError(msg)
        else:
            data_object.data[:, :] -= stray_light_model.data[:, :]
            data_object.uncertainty.array[...] -= stray_light_model.data[:, :]
            data_object.meta["CALSL"] = os.path.basename(str(stray_light_path))
            data_object.meta.history.add_now("LEVEL1-remove_stray_light",
                                             f"stray light removed with {os.path.basename(str(stray_light_model))}")

    logger.info("remove_stray_light finished")
    return data_object
